use serde::{Deserialize, Serialize};
use serde_json::json;
use std::collections::HashMap;

use crate::routes::ApiError;
use crate::routes::item::{SupplyForm, item_does_not_exist_error, variant_does_not_exist_error};
use crate::{get_itemdb, get_pg};
use based::request::api::ToAPI;
use sqlx::FromRow;

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct FlowInfo {
    #[serde(default)]
    pub id: String,
    pub name: String,
    #[serde(default)]
    pub depends: Vec<String>,
    pub next: Option<String>,
    pub produces: Option<Vec<String>>,
}

impl ToAPI for FlowInfo {
    async fn api(&self) -> serde_json::Value {
        json!({
            "id": self.id,
            "name": self.name,
            "depends": self.depends,
            "next": self.next,
            "produces": self.produces
        })
    }
}
/// A production flow
#[derive(Debug, Clone, Serialize, Deserialize, FromRow)]
pub struct Flow {
    /// ID
    pub id: uuid::Uuid,
    /// Tiemstamp when the flow was started
    pub started: chrono::DateTime<chrono::Utc>,
    /// Kind of flow; ID of the describing JSON
    pub kind: String,
    /// Input transactions
    pub input: Option<Vec<uuid::Uuid>>,
    /// Timestamp when the flow was ended
    pub ended: Option<chrono::DateTime<chrono::Utc>>,
    /// The flow succedding this one
    pub next: Option<uuid::Uuid>,
    /// Transactions this flow produced
    pub produced: Option<Vec<uuid::Uuid>>,
}

impl ToAPI for Flow {
    async fn api(&self) -> serde_json::Value {
        let done = if self.ended.is_some() {
            Some(json!({
                "ended": self.ended.map(|x| x.timestamp()),
                "next": self.next,
                "produced": self.produced
            }))
        } else {
            None
        };

        json!({
            "id": self.id,
            "started": self.started.timestamp(),
            "kind": self.kind,
            "input": self.input,
            "done": done
        })
    }
}

impl Flow {
    pub async fn get(id: &uuid::Uuid) -> Option<Self> {
        sqlx::query_as("SELECT * FROM flows WHERE id = $1")
            .bind(id)
            .fetch_optional(get_pg!())
            .await
            .unwrap()
    }

    pub async fn create(kind: &str, input: Option<Vec<uuid::Uuid>>) -> Self {
        sqlx::query_as("INSERT INTO flows (kind, input) VALUES ($1, $2) RETURNING *")
            .bind(kind)
            .bind(input)
            .fetch_one(get_pg!())
            .await
            .unwrap()
    }

    pub async fn end(self) -> Self {
        sqlx::query_as("UPDATE flows SET ended = current_timestamp RETURNING *")
            .fetch_one(get_pg!())
            .await
            .unwrap()
    }

    pub async fn end_with_produce(
        self,
        produced: &[SupplyForm],
    ) -> Result<HashMap<String, Vec<uuid::Uuid>>, ApiError> {
        let mut ret = HashMap::new();
        let mut t_create = Vec::new();
        let mut produced_ref = Vec::with_capacity(ret.len());

        for prod in produced {
            let t = get_itemdb!()
                .get_item(&prod.item)
                .ok_or_else(item_does_not_exist_error)?
                .variant(&prod.variant)
                .ok_or_else(variant_does_not_exist_error)?;

            t_create.push((t, prod));
        }

        for (item, info) in t_create {
            let t = item
                .supply(
                    0.0,
                    Some(&format!("flow::{}::{}", self.kind, self.id)),
                    info.location.as_ref().map(|x| x.as_str()),
                    info.note.as_ref().map(|x| x.as_str()),
                    info.quanta,
                    info.properties.clone(),
                )
                .await;
            ret.entry(item.item_variant_id().clone())
                .or_insert(Vec::new())
                .push(t.id.clone());
            produced_ref.push(t.id);
        }

        sqlx::query("UPDATE transactions SET consumed_timestamp = current_timestamp, produced = $1 WHERE id = $2")
        .bind(produced_ref)
        .bind(self.id)
        .execute(get_pg!()).await.unwrap();

        Ok(ret)
    }

    pub async fn continue_next(self, next_flow: &Flow) -> Self {
        sqlx::query_as("UPDATE transactions SET consumed_timestamp = current_timestamp, \"next\" = $1 WHERE id = $2 RETURNING *")
        .bind(next_flow.id)
        .bind(&self.id)
        .fetch_one(get_pg!()).await.unwrap()
    }
}

/// A note for a Flow
#[derive(Debug, Clone, Serialize, Deserialize, FromRow)]
pub struct FlowNote {
    /// ID
    pub id: uuid::Uuid,
    /// Timestamp when the note was created
    pub time: chrono::DateTime<chrono::Utc>,
    /// Note Content
    pub content: String,
    /// Associated flow
    pub on_flow: uuid::Uuid,
}

impl FlowNote {
    pub async fn create(content: &str, flow: &uuid::Uuid) -> Self {
        sqlx::query_as("INSERT INTO flow_notes (content, on_flow) VALUES ($1, $2) RETURNING *")
            .bind(content)
            .bind(flow)
            .fetch_one(get_pg!())
            .await
            .unwrap()
    }

    pub async fn find_of(id: &str) -> Vec<Self> {
        sqlx::query_as("SELECT * FROM flow_notes WHERE on_flow = $1 ORDER BY time DESC")
            .bind(id)
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }
}

impl ToAPI for FlowNote {
    async fn api(&self) -> serde_json::Value {
        json!({
            "uuid": self.id,
            "timestamp": self.time.timestamp(),
            "content": self.content,
            "on_flow": self.on_flow
        })
    }
}
