mod demand;
pub mod error;
mod location;
mod stat;
mod supply;

use std::str::FromStr;

use based::request::api::ToAPI;
pub use demand::*;
pub use error::*;
pub use location::*;
use rocket::fs::NamedFile;
use rocket::post;
use rocket::serde::json::Json;
use serde::Deserialize;
use serde::Serialize;
pub use stat::*;
pub use supply::*;

use rocket::State;
use rocket::get;
use serde_json::json;

use crate::check_auth;
use crate::config::Config;
use crate::db::ItemDB;
use crate::db::get_items_without_min_satisfied;
use crate::get_locations;
use crate::get_pg;
use crate::routes::Token;
use crate::transaction::Transaction;

use super::api_error;
use crate::routes::FallibleApiResponse;

#[get("/items")]
/// Returns a JSON response with all items in the database.
pub fn get_items_route(itemdb: &State<ItemDB>, t: Token, c: &State<Config>) -> FallibleApiResponse {
    check_auth!(t, c);

    let mut items = Vec::new();
    for item in itemdb.items() {
        items.push(itemdb.get_item(&item).unwrap().api_json());
    }

    Ok(json!({"items": items}))
}

#[get("/item/<item_id>")]
/// Return an API Response for an `Item`
pub fn item_route(
    item_id: &str,
    itemdb: &State<ItemDB>,
    t: Token,
    c: &State<Config>,
) -> FallibleApiResponse {
    check_auth!(t, c);

    let item = itemdb
        .get_item(item_id)
        .ok_or_else(item_does_not_exist_error)?;
    Ok(item.api_json())
}

#[get("/item/<item_id>/image")]
pub async fn item_image_route(item_id: &str, itemdb: &State<ItemDB>) -> Option<NamedFile> {
    let item = itemdb.get_item(item_id)?;

    if let Some(img_path) = &item.image_path {
        return Some(NamedFile::open(img_path).await.ok()?);
    }

    None
}

#[get("/item/<item_id>/variants")]
/// Returns all variants of an Item
pub fn item_variants_page(
    item_id: &str,
    itemdb: &State<ItemDB>,
    t: Token,
    c: &State<Config>,
) -> FallibleApiResponse {
    check_auth!(t, c);

    let item = itemdb
        .get_item(item_id)
        .ok_or_else(item_does_not_exist_error)?;
    let variants = item.get_variants();

    Ok(json!({
        "item": item_id,
        "variants": variants
    }))
}

/// Returns an API Response for a `Transaction`
#[get("/transaction/<transaction>")]
pub async fn transaction_route(
    transaction: &str,
    t: Token,
    c: &State<Config>,
) -> FallibleApiResponse {
    check_auth!(t, c);

    let t = Transaction::get(
        &uuid::Uuid::from_str(&transaction).map_err(|_| api_error("Invalid UUID"))?,
    )
    .await
    .ok_or_else(|| api_error("No transaction with this UUID"))?;
    Ok(t.api().await)
}

#[get("/item/<item_id>/<variant_id>/unique?<field>")]
/// Returns unique values for a field
pub async fn unique_field_route(
    item_id: &str,
    variant_id: &str,
    field: &str,
    itemdb: &State<ItemDB>,
    t: Token,
    c: &State<Config>,
) -> FallibleApiResponse {
    check_auth!(t, c);

    let variant = itemdb
        .get_item(item_id)
        .ok_or_else(item_does_not_exist_error)?
        .variant(variant_id)
        .ok_or_else(variant_does_not_exist_error)?;

    match field {
        "origin" => Ok(json!(
            variant
                .get_unique_origins()
                .await
                .into_iter()
                .filter(|x| !x.starts_with("flow::"))
                .collect::<Vec<_>>()
        )),
        "destination" => Ok(json!(
            variant
                .get_unique_destinations()
                .await
                .into_iter()
                .filter(|x| !x.starts_with("flow::"))
                .collect::<Vec<_>>()
        )),
        _ => Err(api_error("Unknown field")),
    }
}

#[get("/items/expired?<days>")]
/// Get all expired transactions
pub async fn expired_items_route(
    days: Option<&str>,
    t: Token,
    c: &State<Config>,
) -> FallibleApiResponse {
    check_auth!(t, c);

    let days = if let Some(days) = days {
        days.parse()
            .map_err(|_| api_error("Invalid days value"))
            .ok()
    } else {
        None
    };

    let t = Transaction::active_expired(days).await;

    Ok(json!(t))
}

#[get("/items/min")]
/// Return all items without minimum satisfied
pub async fn min_items_route(
    itemdb: &State<ItemDB>,
    t: Token,
    c: &State<Config>,
) -> FallibleApiResponse {
    check_auth!(t, c);

    let t: Vec<_> = get_items_without_min_satisfied(itemdb)
        .await
        .into_iter()
        .map(|x| {
            json!({
                "item_variant": x.0,
                "need": x.1
            }
            )
        })
        .collect();

    Ok(json!(t))
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct MoveTransaction {
    pub to: String,
}

#[post("/transaction/<id>/move", data = "<form>")]
/// Move a transaction to a new location
pub async fn move_transaction_route(id: &str, form: Json<MoveTransaction>) -> FallibleApiResponse {
    let new_loc = &form.to;
    let locations = get_locations!();

    if !locations.contains_key(new_loc) {
        return Err(api_error("No such location"));
    }

    sqlx::query("UPDATE transactions SET location = $1 WHERE id = $2")
        .bind(new_loc)
        .bind(uuid::Uuid::from_str(id).map_err(|_| api_error("Invalid UUID"))?)
        .execute(get_pg!())
        .await
        .unwrap();

    Ok(json!({"ok": 1 }))
}
